from typing import Any, MutableMapping, Callable


def take_last_value(key: str, target: Any, *sources: Any) -> Any:
    """Conflict resolver returning last source
    Args:
        target : current value
        sources : list of update values
    Returns:
        Any: the last value from sources
    """
    try:
        return sources[-1]
    except KeyError:
        return target


def take_first_value(key: str, target: Any, *sources: Any) -> Any:
    """Conflict resolver returning first source
    Args:
        target : current value
        sources : list of update values
    Returns:
        Any: the last value from sources
    """
    try:
        return sources[0]
    except KeyError:
        return target


def merge_values(key: str, target: Any, *sources: Any) -> Any:
    if not isinstance(target, dict) and not key.startswith('$'):
        target = {'$eq': target}
    for s in sources:
        if isinstance(s, MutableMapping):
            target.update(s)
    return target


def dict_deep_update(target: MutableMapping,
                     *sources: MutableMapping,
                     on_conflict: Callable[..., Any] = take_last_value,
                     ) -> MutableMapping:
    """Update a dict with values from others
    Args:
        target : dict to update. (will be modified)
        sources : source dict(s) to update target.
        on_conflict (Callable[[target, *sources], Any], optional):
            define comportment when a key is present in the target
            and/or in multiple source. Defaults to take_last_value.
    Returns:
        MutableMapping: return target so it can be used like this :
        new_dict = dict_deep_update({"key1":"value"}, other)
    """
    merge_keys = []
    for source in sources:
        for k, val in source.items():
            if k in target:
                merge_keys.append(k)
            else:
                target[k] = val
    for k in merge_keys:
        val = target[k]
        src_values = tuple(src[k] for src in sources if k in src)
        if isinstance(val, MutableMapping) and \
                all(isinstance(e, MutableMapping) for e in src_values):
            target[k] = dict_deep_update(val, *src_values,
                                         on_conflict=on_conflict)
        else:
            target[k] = on_conflict(k, val, *src_values)
    return target



from typing import Optional

from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from contextlib import asynccontextmanager


class Connection:
    client: AsyncIOMotorClient = None
    database: AsyncIOMotorDatabase = None

    async def connect(self, *args, **kwargs):
        print(f'Connection to {args} {kwargs}')
        self.client = AsyncIOMotorClient(*args, **kwargs)
        self.database = self.client.get_default_database()

    async def disconnect(self) -> None:
        self.client = None
        self.database = None


@asynccontextmanager
async def use_client(
    *args,
    client: Optional[AsyncIOMotorClient] = None,
    **kwargs
):
    if not client:
        client = AsyncIOMotorClient(*args, **kwargs)
    old_client = connection.client
    old_database = connection.database
    connection.client = client
    connection.database = client.get_default_database()
    yield
    connection.client = old_client
    connection.database = old_database


connection = Connection()



from typing import (
    Any, Generator, List, Callable, Dict, Optional, AsyncGenerator, Tuple, Union, Type
)
from motor.motor_asyncio import (
    AsyncIOMotorCollection, AsyncIOMotorDatabase, AsyncIOMotorCursor,
    AsyncIOMotorClientSession
)
from pymongo import ASCENDING, DESCENDING
from pymongo.results import InsertOneResult, UpdateResult
from motorized.query import Q, QueryDict
from motorized.client import connection
from motorized.exceptions import NotConnectedException
from contextlib import contextmanager


class QuerySet:
    def __init__(self, model, initial_query: Optional[Q] = None):
        self.model = model
        self._query = initial_query or Q()
        self._initial_query: Optional[Q] = initial_query
        self._limit = None
        self._sort = None
        self.database: Optional[AsyncIOMotorDatabase] = None
        self._session: Optional[AsyncIOMotorClientSession] = None
        # _collection_name allow you to override used collection name,
        # priority order is: _collection_name > Document.Mongo.collection
        self._collection_name: Optional[str] = None

    def copy(self) -> "QuerySet":
        instance = self.__class__(self.model)
        instance._query = self._query.copy()
        instance._limit = self._limit
        instance._sort = self._sort
        instance._collection_name = self._collection_name
        instance.use_database(self.database)
        return instance

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.model.__name__}: ' \
               f'{self._query}>'

    @classmethod
    def from_query(cls,
                   model: Type["Document"],  # noqa: F821
                   query: Q) -> "QuerySet":
        instance = QuerySet(model)
        instance._query = query
        return instance

    def use_database(self, database: Optional[AsyncIOMotorDatabase]) -> None:
        """if `database` is None then the default database will be used.
        """
        self.database = database

    def use_session(self, session: AsyncIOMotorClientSession) -> "QuerySet":
        instance = self.copy()
        instance._session = session
        return instance

    @contextmanager
    def collection_name(self, collection_name: str) -> Generator[None, None, "QuerySet"]:
        """Allow the user to override the destination collection,
        use this context manager to copy items from one collection to an other.
        usage:
        ```python
        item = await Document.objects.first()
        with Document.objects.collection_name("migration"):
            await item.save()
        ```
        """
        original_collection_name = self._collection_name
        self._collection_name = collection_name
        yield self
        self._collection_name = original_collection_name

    async def __aiter__(self) -> AsyncGenerator["Document", None]:  # noqa: F821,E501
        cursor = await self.find()
        async for data in cursor:
            instance = self.model(**data)
            yield instance

    async def create(self, *args, **kwargs) -> "Document":  # noqa: F821
        instance = self.model(*args, **kwargs)
        await instance.save()
        return instance

    async def insert_one(self, data: Dict, **kwargs) -> InsertOneResult:
        kwargs.setdefault('session', self._session)
        return await self.collection.insert_one(data, **kwargs)

    @property
    def collection(self) -> AsyncIOMotorCollection:
        collection_name: str = self._collection_name or self.model.Mongo.collection
        if self.database is not None:
            return getattr(self.database, collection_name)
        if not connection.database:
            raise NotConnectedException(
                'You need to use connection.connect before using collection'
            )
        return getattr(connection.database, collection_name)

    async def count(self, **kwargs) -> int:
        return await self.collection.count_documents(
            self._query.query, **kwargs)

    async def first(self) -> Optional["Document"]:  # noqa: F821
        raw_data = await self.find_one()
        if raw_data is None:
            return None
        instance = self.model(**raw_data)
        return instance

    async def drop(self) -> None:
        await self.collection.drop()

    def filter(self, query: Optional[Q] = None, /, **kwargs) -> "QuerySet":
        instance = self.copy()
        if query:
            instance._query += query
        instance._query += Q(**kwargs)
        return instance

    def exclude(self, **kwargs) -> "QuerySet":
        instance = self.copy()
        inner_query = Q()
        inner_query.query = Q.convert_kwargs_to_query(**kwargs, invert=True)
        instance._query += inner_query
        return instance

    def limit(self, size: int) -> "QuerySet":
        instance = self.copy()
        instance._limit = size
        return instance

    def order_by(
        self,
        ordering: Optional[Union[str, List[str]]]
    ) -> "QuerySet":
        instance = self.copy()
        if ordering:
            instance._sort = self._sort_instruction(
                ordering if hasattr(ordering, '__iter__') else [ordering]
            )
        else:
            instance._sort = None
        return instance

    def __add__(self, other: "QuerySet") -> "QuerySet":
        instance = self.copy()
        instance._query += other._query
        return instance

    async def all(self) -> List["Document"]:  # noqa: F821
        return [instance async for instance in self]

    async def map(self, func: Callable) -> List[Any]:
        """Apply `func` to all match in the query queryset and return the
        result of the function in a list.
        """
        return list([await func(instance) async for instance in self])

    async def distinct(self, key: str, **kwargs) -> List[Any]:
        if not self._query.is_empty():
            kwargs.setdefault('filter', self._query.query)
        return await self.collection.distinct(key, **kwargs)

    async def aggregate(self, pipeline, **kwargs) -> Any:
        kwargs.setdefault('session', self._session)
        return await self.collection.aggregate(pipeline, **kwargs)

    async def find(self, **kwargs) -> AsyncIOMotorCursor:
        kwargs.setdefault('session', self._session)
        cursor = self.collection.find(self._query.query, **kwargs)
        return await self._paginate_cursor(cursor)

    async def find_one(self, **kwargs) -> Dict:
        kwargs.setdefault('session', self._session)
        return await self.collection.find_one(self._query.query, **kwargs)

    async def get(self, **kwargs) -> "Document":  # noqa: F821
        instance = self.filter(**kwargs)
        cursor = await instance.find()
        try:
            first = await cursor.__anext__()
        except StopAsyncIteration as error:
            raise self.model.DocumentNotFound(self._query.query) from error
        try:
            await cursor.__anext__()
        except StopAsyncIteration:
            return self.model(**first)

        raise self.model.TooManyMatchException

    def fresh(self) -> "QuerySet":
        """Return a fresh queryset without any filtering/ordering/limiting
        parameter as fresh as new.
        """
        instance = QuerySet(self.model, self._initial_query)
        instance.use_database(self.database)
        return instance

    async def values_list(self, fields: List[str], flat=False, noid=False):
        if isinstance(fields, str):
            fields = (fields,)
        projection = {f: True for f in fields}
        if noid:
            projection['_id'] = False

        cursor: AsyncIOMotorCursor = await self.find(projection=projection)
        if not flat:
            return list([item async for item in cursor])
        assert len(fields) == 1, 'You can only have one field using flat=True'
        field_name = fields[0]
        return list([value[field_name] async for value in cursor])

    async def _paginate_cursor(
        self,
        cursor: AsyncIOMotorCursor
    ) -> AsyncIOMotorCursor:
        if self._sort:
            cursor = cursor.sort(self._sort)
        if self._limit:
            cursor = cursor.limit(self._limit)
        return cursor

    @staticmethod
    def _sort_instruction(order: List[str]) -> List[Tuple[str, int]]:
        """Convert a list for format:
        ['name', '-age'] to:
        [('name': 1), ('age': -1)]
        """
        def generate_tuple(word) -> Tuple[str, int]:
            if word.startswith('-'):
                return (word[1:], DESCENDING)
            return (word, ASCENDING)

        return [generate_tuple(word) for word in order]

    async def exists(self) -> bool:
        return await self.collection.count_documents(
            self._query.query, limit=1) > 0

    async def delete(self, **kwargs):
        kwargs.setdefault('session', self._session)
        return await self.collection.delete_many(self._query.query, **kwargs)

    async def delete_one(self, **kwargs):
        kwargs.setdefault('session', self._session)
        return await self.collection.delete_one(self._query.query, **kwargs)

    async def pop(
        self,
        session: Optional[AsyncIOMotorClientSession] = None,
        **kwargs
    ) -> "Document":  # noqa: F821
        """Retrieve a document and delete it from the database.
        """
        instance = self.filter(**kwargs)
        document_dict = await instance.collection.find_one_and_delete(
            instance.query._query,
            session=session or self._session
        )
        document = self.model(**document_dict)
        return document

    async def indexes(self):
        return [index async for index in self.collection.list_indexes()]

    async def _aggregate(
        self,
        operator: str,
        fields: Union[str, List[str]],
        **kwargs
    ) -> Union[int, Dict]:
        """Create a mongodb pipeline on all given `fields` using the operator
        (ex: $sum) if the `fields` parameter is a List then a dictionary of the
        values will be returned, the keys will be the fields names,
        if the `fields` parameter is an instance of string then the value will
        be directly returned.
        """
        kwargs.setdefault('session', self._session)
        if isinstance(fields, str):
            fields = [fields]
        pipeline = [
            {"$match": self._query.query},
            {
                "$group": {
                    "_id": "total",
                    **{field: {operator: f'${field}'} for field in fields}
                }
            }
        ]
        cursor = self.collection.aggregate(pipeline, **kwargs)
        result = await cursor.next()
        # if the list only has one element we just return the first result
        if len(fields) == 1:
            return result[fields[0]]
        result.pop('_id', None)
        return result

    async def sum(self, fields: Union[str, List[str]]) -> Union[int, Dict]:
        return await self._aggregate('$sum', fields)

    async def avg(self, fields: Union[str, List[str]]) -> Union[float, Dict]:
        return await self._aggregate('$avg', fields)

    async def update(self, **kwargs) -> UpdateResult:
        updater = QueryDict(**kwargs)
        return await self.collection.update_many(
            self._query.query,
            {"$set": updater},
            session=self._session
        )

    async def unset(self, fields_names: List[str]) -> UpdateResult:
        """Remove fields from documents
        example:
        ```python
        await Book.objects.filter(some_field__exists=True).unset(['some_field'])
        ```
        Note this could break the validation of the ODM and lead to data not
        being consitent anymore, use it carefully
        """
        fields = list(['.'.join(item.split('__')) for item in fields_names])
        fields_dict = {field: "" for field in fields}
        return await self.collection.update_many(
            self._query.query,
            {'$unset': fields_dict},
            session=self._session
        )



from contextlib import contextmanager
from typing import Optional, Union, Any, Dict, Type, List, Generator
from pydantic import BaseModel, Field, validate_model
from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass
from pymongo.results import InsertOneResult, UpdateResult

from motorized.queryset import QuerySet
from motorized.query import Q
from motorized.types import PydanticObjectId, ObjectId
from motorized.exceptions import DocumentNotSavedError, MotorizedError


@contextmanager
def hide_fields(instance: BaseModel, *fields):
    """Hide fields from pydantic,
    use carefully.
    """
    original_annotations = instance.__annotations__
    original_fields = instance.__fields__
    new_fields = original_fields.copy()
    new_annotations = original_annotations.copy()
    for field in fields:
        new_fields.pop(field, None)
        new_annotations.pop(field, None)
    instance.__fields__ = new_fields
    yield instance
    instance.__fields__ = original_fields
    instance.__annotations__ = original_annotations


class DocumentMeta(ModelMetaclass):
    def __new__(cls, name, bases, optdict: Dict) -> Type['Document']:
        # optdict.pop('objects', None)
        # optdict.pop('__annotations__', {}).pop('objects', None)
        instance: Type[Document] = super().__new__(cls, name, bases, optdict)
        if name not in ('Document',):
            cls._populate_default_mongo_options(
                cls, name, instance, optdict.get('Mongo'))

        class DocumentError(MotorizedError):
            pass

        class TooManyMatchException(DocumentError):
            pass

        class DocumentNotFound(DocumentError):
            pass

        instance.DocumentError = DocumentError
        instance.TooManyMatchException = TooManyMatchException
        instance.DocumentNotFound = DocumentNotFound
        instance.objects = instance.Mongo.manager_class(
            instance, getattr(instance.Mongo, 'filters', None))
        return instance

    def _populate_default_mongo_options(cls, name: str, instance: "Document",
                                        custom_mongo_settings_class) -> None:
        class Mongo:
            pass

        # forbid re-utilisation of the Mongo class between
        # inheritance of the class
        try:
            if instance.Mongo.class_name != name:
                instance.Mongo = Mongo()
        except AttributeError:
            pass

        if custom_mongo_settings_class:
            instance.Mongo = custom_mongo_settings_class

        default_settings = {
            'collection': name.lower() + 's',
            'manager_class': QuerySet,
            'local_fields': [],
            'class_name': name,
            'filters': None,
        }

        for attribute_name, default_value in default_settings.items():
            if not hasattr(instance.Mongo, attribute_name):
                setattr(instance.Mongo, attribute_name, default_value)


class Document(BaseModel, metaclass=DocumentMeta):
    id: Optional[PydanticObjectId] = Field(alias='_id')

    class Config:
        json_encoders = {ObjectId: str}
        validate_assignment = True

    class Mongo:
        manager_class = QuerySet

    def __init__(self, *args, **kwargs) -> None:
        BaseModel.__init__(self, *args, **self._transform(**kwargs))

    def get_query(self) -> Q:
        document_id = getattr(self, 'id', None)
        if not document_id:
            raise DocumentNotSavedError('document has no id.')
        return Q(_id=document_id)

    async def _create_in_db(self, creation_dict: Dict) -> InsertOneResult:
        response = await self.objects.insert_one(creation_dict)
        self.id = response.inserted_id
        return response

    async def _update_in_db(self, update_dict: Dict) -> UpdateResult:
        return await self.objects.collection.update_one(
            filter={'_id': self.id},
            update={'$set': update_dict}
        )

    @classmethod
    def _is_field_to_save(cls, field_name: str) -> bool:
        return (
            not field_name.startswith('_') and
            field_name not in cls.Mongo.local_fields and
            field_name in cls.__fields__
        )

    async def to_mongo(self) -> Dict:
        """Convert the current model dictionary to database output dict,
        this also mean the aliased fields will be stored in the alias name
        instead of their name in the document declaration.
        """
        saving_data = self.dict()

        # remove any field that is not to save, this has to be done
        # before the aliasing resolving to allow to save/load fields
        # that starts with _
        saving_data = dict(
            {
                k: v for k, v in saving_data.items()
                if self._is_field_to_save(k)
            }
        )

        # resolve all alised fields to be saved in their alias name
        for field in self._aliased_fields():
            saving_data[field.alias] = saving_data.pop(field.name, None)

        return saving_data

    async def save(self) -> Union[InsertOneResult, UpdateResult]:
        data = await self.to_mongo()
        document_id = data.pop('_id', None)
        if document_id is None:
            return await self._create_in_db(data)
        return await self._update_in_db(data)

    async def commit(self) -> "Document":
        """Same as `.save` but return the current instance.
        """
        await self.save()
        return self

    async def delete(self) -> "Document":
        """Delete the current instance from the database,
        to the deleted the instance need to have a .id set, in any case the
        function will return the instance itself
        """
        try:
            await self.objects.filter(self.get_query()).delete_one()
            # qs = self.objects.from_query(self, self.get_query())
            # await qs.delete_one()
        except DocumentNotSavedError:
            pass
        setattr(self, 'id', None)
        return self

    async def fetch(self) -> "Document":
        """Return a fresh instance of the current document from the database.
        """
        return await self.objects.filter(self.get_query()).get()

    @classmethod
    def _aliased_fields(cls) -> Generator[List[ModelField], None, None]:
        """Return the list of fields with aliases
        """
        return [
            field for field in cls.__fields__.values()
            if field.name != field.alias
        ]

    def _transform(self, **kwargs) -> Dict:
        """Override this method to change the input database before having it
        being validated/parsed by BaseModel (pydantic)
        """
        return kwargs

    async def reload(self) -> "Document":
        # fetch an validate input data from database
        model_data = await self.objects.filter(self.get_query()).find_one()
        model_data.pop('_id')
        return self.update(model_data)

    def update(self, input_data: Dict) -> "Document":
        """Update the current instance with the given `input_data` after
        validation return the object itself (without saving it in the database)
        """
        validate_model(self, input_data)
        allow_extra: bool = getattr(self.Config, 'extra', 'ignore') == 'allow'

        # load the fields into the current instance
        for field, value in input_data.items():
            if allow_extra or hasattr(self, field):
                setattr(self, field, value)
        return self

    def __repr__(self):
        def get_field_entry(field: Field) -> str:
            return f"{field.name}={getattr(self, field.name)}"

        fields = ", ".join([
            get_field_entry(field)
            for field in self.__fields__.values()
            if field.name not in self.Mongo.local_fields
        ])
        return f'{self.__class__.__name__}({fields})'

    def __setattr__(self, name: str, value: Any) -> None:
        # allow any private attribute to be passed
        if name.startswith('_') or name in self.Mongo.local_fields:
            return object.__setattr__(self, name, value)

        # otherwise we let pydantic decide
        return super().__setattr__(name, value)

    def _has_method(self, name: str) -> bool:
        return callable(getattr(self, name, None))



from bson import ObjectId
from bson.errors import InvalidId


class PydanticObjectId(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, ObjectId):
            raise TypeError('ObjectId required')
        return v

    @classmethod
    def __modify_schema__(cls, schema: dict):
        schema['type'] = 'string'


class InputObjectId(str):
    """Represent a string but will be casted as an ObjectId
    this should be used when retriving an ObjectId from a user,
    ex: fastapi endpoint parameter
    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        try:
            return ObjectId(str(v))
        except (ValueError, InvalidId):
            raise TypeError('ObjectId required')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['motorized']

package_data = \
{'': ['*']}

install_requires = \
['motor>=2.5.1,<3.0.0', 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'motorized',
    'version': '0.3.3',
    'description': 'Asyncio mongodb ODM',
    'long_description': None,
    'author': 'Sebastien Nicolet',
    'author_email': 'snicolet95@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
